# tex
# text.usetex : True

# figure size
figure.figsize : 6, 4

# colors
text.color        : black
figure.facecolor  : white
xtick.color       : black
ytick.color       : black
axes.facecolor    : white
axes.edgecolor    : black
axes.labelcolor   : black
savefig.facecolor : white

# font sizes
font.size        : 11
axes.titlesize   : 11
axes.labelsize   : 11
xtick.labelsize  : 11
ytick.labelsize  : 11
legend.fontsize  : 11
figure.titlesize : 11

# padding
xtick.major.pad : 6
xtick.minor.pad : 6
ytick.major.pad : 6
ytick.minor.pad : 6
axes.titlepad   : 8
axes.labelpad   : 8
